#ifndef LIB94_LIB94_HPP
#define LIB94_LIB94_HPP

#include <filesystem>
#include <optional>
#include <variant>
#include <string>
#include <random>
#include <vector>
#include <deque>
#include <set>

#ifndef LIB94_CORE_SIZE
#define LIB94_CORE_SIZE 8000
#endif

namespace lib94 {

  typedef int_least32_t number_t;

  enum opcode : uint8_t {
    DAT, MOV, ADD, SUB,
    MUL, DIV, MOD, JMP,
    JMZ, JMN, DJN, SEQ,
    SNE, SLT, SPL, NOP
  };

  enum modifier : uint8_t {
    A, B, AB, BA,
    F, X, I
  };

  enum mode : uint8_t {
    IMMEDIATE, DIRECT,
    A_INDIRECT, B_INDIRECT,
    A_DECREMENT, B_DECREMENT,
    A_INCREMENT, B_INCREMENT
  };

  struct instruction {
    opcode op;
    modifier mod;
    mode amode;
    mode bmode;
    number_t anumber;
    number_t bnumber;
  };

  struct warrior {
    std::string name;
    std::string author;

    number_t org;

    std::vector<instruction> instructions;
  };

  void seed_prng(uint_fast64_t seed);

  std::string instruction_to_string(const instruction &instr);

  std::variant<warrior *, std::string> compile_warrior(std::string source);
  bool save_warrior(const warrior &w, const std::filesystem::path &to);
  std::optional<warrior *> load_warrior(const std::filesystem::path &from);

  void clear_core(const instruction &background);
  void clear_core_random();

  //warrior pointers need to remain valid for other
  //functions to return valid things during the round
  bool init_round(const warrior *const *warriors, size_t count);

  size_t alive_warrior_count();

  //asserts that there is a next warrior
  const warrior *get_next_warrior();
  const std::deque<number_t> &get_processes(const warrior *for_warrior);
  //asserts that there is a next process
  number_t get_next_process(const warrior *for_warrior);

  const std::set<number_t> &get_written_addresses();
  const std::set<number_t> &get_read_addresses();
  const std::set<number_t> &get_executed_addresses();
  void clear_address_sets();

  const instruction &get_instruction(number_t address);

  //assumes that there is a next warrior
  //returns a warrior if it dies
  const warrior *single_step();

}

#endif
